# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Power distribution algorithm to distribute power between batteries."""

import logging
import math
from dataclasses import dataclass
from typing import NamedTuple, Sequence

from frequenz.client.microgrid import BatteryData, InverterData

from ...._internal._math import is_close_to_zero
from ..result import PowerBounds

_logger = logging.getLogger(__name__)


@dataclass()
class AggregatedBatteryData:
    """Aggregated battery data."""

    component_id: int
    """The component ID of the first battery.

    This is only used to identify the pair of battery and inverter.
    """

    soc: float
    """The aggregated SoC of the batteries."""

    capacity: float
    """The aggregated capacity of the batteries."""

    soc_upper_bound: float
    """The aggregated upper SoC bound of the batteries."""

    soc_lower_bound: float
    """The aggregated lower SoC bound of the batteries."""

    power_bounds: PowerBounds
    """The aggregated power bounds of the batteries."""

    def __init__(self, batteries: list[BatteryData]) -> None:
        """Create DistBatteryData from BatteryData.

        Aggregates the data of the batteries:

        * Capacity: Sum of capacities of all batteries.
        * SoC: Weighted average of SoCs of all batteries.
        * SoC bounds: Weighted average of SoC bounds of all batteries.
        * Power inclusion bounds: Sum of power inclusion bounds of all batteries.
        * Power exclusion bounds: Largest power exclusion bound multiplied by
            the number of batteries.

        Args:
            batteries: The batteries to aggregate.
        """
        assert len(batteries) > 0, "AggregatedBatteryData: No batteries given."

        # We need only one component ID for DistBatteryData to be able to
        # identify the pair
        self.component_id = batteries[0].component_id

        self.capacity = sum(b.capacity for b in batteries)

        if self.capacity != 0.0:
            self.soc = sum(b.soc * b.capacity for b in batteries) / self.capacity
            self.soc_upper_bound = (
                sum(b.soc_upper_bound * b.capacity for b in batteries) / self.capacity
            )
            self.soc_lower_bound = (
                sum(b.soc_lower_bound * b.capacity for b in batteries) / self.capacity
            )
        else:
            self.soc = math.nan
            self.soc_upper_bound = math.nan
            self.soc_lower_bound = math.nan

        self.power_bounds = _aggregate_battery_power_bounds(
            list(
                map(
                    lambda metrics: PowerBounds(
                        inclusion_upper=metrics.power_inclusion_upper_bound,
                        inclusion_lower=metrics.power_inclusion_lower_bound,
                        exclusion_upper=metrics.power_exclusion_upper_bound,
                        exclusion_lower=metrics.power_exclusion_lower_bound,
                    ),
                    batteries,
                )
            )
        )


def _aggregate_battery_power_bounds(
    battery_metrics: Sequence[PowerBounds],
) -> PowerBounds:
    """Calculate bounds for a set of batteries located behind one set of inverters.

    Args:
        battery_metrics: List of PowerBounds for each battery.

    Returns:
        A PowerBounds object containing the aggregated bounds for all given batteries
    """
    assert len(battery_metrics) > 0, "No batteries given."

    # Calculate the aggregated bounds for the set of batteries
    power_inclusion_upper_bound = sum(
        bounds.inclusion_upper for bounds in battery_metrics
    )
    power_inclusion_lower_bound = sum(
        bounds.inclusion_lower for bounds in battery_metrics
    )

    # To satisfy the largest exclusion bounds in the set we need to
    # provide the power defined by the largest bounds multiplied by the
    # number of batteries in the set.
    power_exclusion_upper_bound = max(
        bounds.exclusion_upper for bounds in battery_metrics
    ) * len(battery_metrics)
    power_exclusion_lower_bound = min(
        bounds.exclusion_lower for bounds in battery_metrics
    ) * len(battery_metrics)

    return PowerBounds(
        inclusion_lower=power_inclusion_lower_bound,
        exclusion_lower=power_exclusion_lower_bound,
        exclusion_upper=power_exclusion_upper_bound,
        inclusion_upper=power_inclusion_upper_bound,
    )


class InvBatPair(NamedTuple):
    """InvBatPair with inverter and adjacent battery data."""

    battery: AggregatedBatteryData
    """The battery data."""

    inverter: list[InverterData]
    """The inverter data."""


@dataclass
class AvailabilityRatio:
    """Availability ratio for a battery-inverter pair."""

    battery_id: int
    """The battery ID."""

    inverter_ids: list[int]
    """The inverter IDs."""

    ratio: float
    """The availability ratio."""

    min_power: float
    """The minimum power that can be set for the battery-inverters pair."""


@dataclass
class _Power:
    """Helper class for distribution algorithm."""

    upper_bound: float
    """The upper bound of the power that can be set for the battery."""

    power: float
    """The power to be set for the inverter."""


_InverterSet = frozenset[int]
"""A set of inverter IDs."""


@dataclass
class _Allocation:
    """Helper class for distribution algorithm."""

    inverter_ids: _InverterSet
    """The IDs of the inverters."""

    power: float
    """The power to be set for the inverters."""


@dataclass
class DistributionResult:
    """Distribution result."""

    distribution: dict[int, float]
    """The power to be set for each inverter.

    The key is inverter ID, and the value is the power that should be set for
    that inverter.
    """

    remaining_power: float
    """The power which could not be distributed because of bounds."""


class BatteryDistributionAlgorithm:
    r"""Distribute power between many components.

    The purpose of this tool is to keep equal SoC level in the batteries.
    It takes total power that should be to be set for some subset of battery-inverter
    pairs. The total power is distributed between given battery-inverter pairs.
    Distribution is calculated based on data below:

    * Battery current SoC.
    * Battery upper and lower SoC bound.
    * Battery capacity.
    * Battery lower and upper power bound.
    * Inverter lower and upper active power bound.

    # Distribution algorithm

    Lets assume that:

    * `N` - number of batteries
    * `power_w` - power to distribute
    * `capacity[i]` - capacity of i'th battery
    * `available_soc[i]` - how much SoC remained to reach:
        * SoC upper bound - if need to distribute power that charges inverters.
        * SoC lower bound - if need to distribute power that discharges inverters.
        * `0` - if SoC is outside SoC bounds.

    * `total_capacity` - `sum(c for c in capacity.values())`
    * `capacity_ratio[i]` - `capacity[i]/total_capacity`


    We would like our distribution to meet the equation:

    ```
    distribution[i] = power_w * capacity_ratio[i] * x[i]
    ```

    where:

    ```
    sum(capacity_ratio[i] * x[i] for i in range(N)) == 1
    ```

    Let `y` be our unknown, the proportion to discharge each battery would be
    (1):

    ```
    x[i] = available_soc[i]*y
    ```

    We can compute `y` from equation above (2):

    ```
    sum(capacity_ratio[i] * x[i] for i in range(N)) == 1
    # =>
    sum(capacity_ratio[i] * available_soc[i] * y for i in range(N)) == 1
    # =>
    y = 1 / sum(capacity_ratio[i] * available_soc[i])
    ```

    Now we know everything and we can compute distribution:

    ```
    distribution[i] = power_w * capacity_ratio[i] * x[i]  # from (1)
    distribution[i] = \
            power_w * capacity_ratio[i] * available_soc[i] * y  # from (2)
    distribution[i] = power_w * capacity_ratio[i] * available_soc[i] * \
            1/sum(capacity_ratio[i] * available_soc[i])
    ```

    Let:

    ```
    battery_availability_ratio[i] = capacity_ratio[i] * available_soc[i]
    total_battery_availability_ratio = sum(battery_availability_ratio)
    ```

    Then:
    ```
    distribution[i] = power_w * battery_availability_ratio[i] \
            / total_battery_availability_ratio
    ```
    """

    def __init__(self, distributor_exponent: float = 1) -> None:
        """Create distribution algorithm instance.

        Args:
            distributor_exponent: How fast the batteries should strive to the
                equal SoC level. Should be float >= 0. Defaults=1.
                For example for distributor_exponent equal:
                    * 1 - means that proportion will be linear from SoC.
                    * 2 - means proportion would be like squared from SoC
                    * 3 - means proportion would be like x^3 from SoC.

        Example:
            Lets say we have two batteries `Bat1` and `Bat2`. All parameters
            except SoC are equal. SoC bounds for each battery is `lower = 20`,
            `upper = 80`.

            # Example 1

            Let:

            * `Bat1.soc = 70` and `Bat2.soc = 50`.
            * `Bat1.available_soc = 10`, `Bat2.available_soc = 30`
            * `Bat1.available_soc / Bat2.available_soc = 3`

            A request power of 8000W will be distributed as follows, for different
            values of `distribution_exponent`:

            | distribution_exponent | Bat1 | Bat2 |
            |-----------------------|------|------|
            | 0                     | 4000 | 4000 |
            | 1                     | 2000 | 6000 |
            | 2                     | 800  | 7200 |
            | 3                     | 285  | 7715 |


            # Example 2

            Let:

            * `Bat1.soc = 50` and `Bat2.soc = 20`.
            * `Bat1.available_soc = 30`, `Bat2.available_soc = 60`
            * `Bat1.available_soc / Bat2.available_soc = 2`

            A request power of 900W will be distributed as follows, for different
            values of `distribution_exponent`.

            | distribution_exponent | Bat1 | Bat2 |
            |-----------------------|------|------|
            | 0                     | 450  | 450  |
            | 1                     | 300  | 600  |
            | 2                     | 180  | 720  |
            | 3                     | 100  | 800  |

            # Example 3

            Let:

            * `Bat1.soc = 44` and `Bat2.soc = 64`.
            * `Bat1.available_soc = 36 (80 - 44)`, `Bat2.available_soc = 16 (80 - 64)`

            A request power of 900W will be distributed as follows, for these values of
            `distribution_exponent`:

            If `distribution_exponent` is:

            | distribution_exponent | Bat1 | Bat2 |
            |-----------------------|------|------|
            | 0                     | 450  | 450  |
            | 0.5                   | 600  | 400  |

        Raises:
            ValueError: If distributor_exponent < 0

        """
        super().__init__()

        if distributor_exponent < 0:
            raise ValueError("Distribution factor should be float >= 0.")
        self._distributor_exponent: float = distributor_exponent

    def _total_capacity(self, components: list[InvBatPair]) -> float:
        """Sum capacity between all batteries in the components list.

        Args:
            components: list of the components

        Raises:
            ValueError: If total capacity is 0.

        Returns:
            Sum of all batteries capacity in the components list.
        """
        total_capacity: float = sum(bat.capacity for bat, _ in components)
        if is_close_to_zero(total_capacity):
            msg = "All batteries have capacity 0."
            _logger.error(msg)
            raise ValueError(msg)

        return total_capacity

    def _compute_battery_availability_ratio(
        self,
        components: list[InvBatPair],
        available_soc: dict[int, float],
        excl_bounds: dict[int, float],
    ) -> tuple[list[AvailabilityRatio], float]:
        r"""Compute battery ratio and the total sum of all of them.

        battery_availability_ratio = capacity_ratio[i] * available_soc[i]
        Where:
        capacity_ratio[i] = components[i].battery.capacity \
            / sum(battery.capacity for battery, _ in components)

        Args:
            components: list of the components
            available_soc: How much SoC remained to reach
                * SoC upper bound - if need to distribute consumption power
                * SoC lower bound - if need to distribute supply power
            excl_bounds: Exclusion bounds for each inverter

        Returns:
            Tuple where first argument is battery availability ratio for each
                battery-inverter pair. The list is sorted by ratio in
                descending order.  The second element of the tuple is total sum
                of all battery ratios in the list.
        """
        total_capacity = self._total_capacity(components)
        battery_availability_ratio: list[AvailabilityRatio] = []
        total_battery_availability_ratio: float = 0.0

        for pair in components:
            battery, inverters = pair
            capacity_ratio = battery.capacity / total_capacity
            soc_factor: float = pow(
                available_soc[battery.component_id], self._distributor_exponent
            )

            ratio = capacity_ratio * soc_factor

            inverter_ids = [inv.component_id for inv in inverters]
            inverter_ids.sort(key=lambda item: (excl_bounds[item], item), reverse=True)

            battery_availability_ratio.append(
                AvailabilityRatio(
                    battery.component_id,
                    inverter_ids,
                    ratio,
                    # Min power we need to request from the pair.
                    # Note that indvidual inverters may have lower min power
                    # and need to be checked individually.
                    min_power=max(
                        excl_bounds[battery.component_id],
                        min(excl_bounds[inverter_id] for inverter_id in inverter_ids),
                    ),
                )
            )

            total_battery_availability_ratio += ratio

        battery_availability_ratio.sort(
            key=lambda item: (item.min_power, item.ratio), reverse=True
        )

        return battery_availability_ratio, total_battery_availability_ratio

    def _distribute_power(  # pylint: disable=too-many-arguments
        self,
        *,
        components: list[InvBatPair],
        power_w: float,
        available_soc: dict[int, float],
        incl_bounds: dict[int, float],
        excl_bounds: dict[int, float],
    ) -> DistributionResult:
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        """Distribute power between given components.

        After this method power should be distributed between batteries
        in a way that equalize SoC between batteries.

        Args:
            components: list of components.
            power_w: power to distribute
            available_soc: how much SoC remained to reach:
                * SoC upper bound - if need to distribute consumption power
                * SoC lower bound - if need to distribute supply power
            incl_bounds: Inclusion bounds for each inverter
            excl_bounds: Exclusion bounds for each inverter

        Returns:
            Distribution result.
        """
        (
            battery_availability_ratio,
            sum_ratio,
        ) = self._compute_battery_availability_ratio(
            components, available_soc, excl_bounds
        )

        # sum_ratio == 0 means that all batteries are fully charged / discharged
        if is_close_to_zero(sum_ratio):
            final_distribution = {
                inverter.component_id: 0.0
                for _, inverters in components
                for inverter in inverters
            }
            return DistributionResult(final_distribution, power_w)

        # key: inverter_ids, value: _Power(upper_bound, power)
        distribution: dict[_InverterSet, _Power] = {}
        distributed_power: float = 0.0
        reserved_power: float = 0.0
        power_to_distribute: float = power_w
        used_ratio: float = 0.0
        ratio = sum_ratio
        excess_reserved: dict[_InverterSet, float] = {}
        deficits: dict[_InverterSet, float] = {}

        for ratio_data in battery_availability_ratio:
            inverter_set = _InverterSet(ratio_data.inverter_ids)
            # ratio = 0, means all remaining batteries reach max SoC lvl or have no
            # capacity
            if is_close_to_zero(ratio):
                distribution[inverter_set] = _Power(
                    upper_bound=0.0,
                    power=0.0,
                )
                continue

            power_to_distribute = power_w - reserved_power
            calculated_power = power_to_distribute * ratio_data.ratio / ratio
            reserved_power += max(calculated_power, ratio_data.min_power)
            used_ratio += ratio_data.ratio
            ratio = sum_ratio - used_ratio

            # If the power allocated for that inverter set is out of bound,
            # then we need to distribute more power over all remaining batteries.
            incl_bound = min(
                sum(
                    incl_bounds[inverter_id] for inverter_id in ratio_data.inverter_ids
                ),
                incl_bounds[ratio_data.battery_id],
            )
            if calculated_power > incl_bound:
                excess_reserved[inverter_set] = incl_bound - ratio_data.min_power
            # # Distribute between remaining batteries
            elif calculated_power < ratio_data.min_power:
                deficits[inverter_set] = calculated_power - ratio_data.min_power
            else:
                excess_reserved[inverter_set] = calculated_power - ratio_data.min_power

            distributed_power += ratio_data.min_power
            distribution[inverter_set] = _Power(
                upper_bound=incl_bound,
                power=ratio_data.min_power,
            )

        for inverter_ids, deficit in deficits.items():
            while not is_close_to_zero(deficit) and deficit < 0.0:
                if not excess_reserved:
                    break
                largest = _Allocation(
                    *max(excess_reserved.items(), key=lambda item: item[1])
                )

                if is_close_to_zero(largest.power) or largest.power < 0.0:
                    break
                if largest.power >= -deficit or math.isclose(largest.power, -deficit):
                    excess_reserved[largest.inverter_ids] += deficit
                    deficits[inverter_ids] = 0.0
                    deficit = 0.0
                else:
                    deficit += excess_reserved[largest.inverter_ids]
                    deficits[inverter_ids] = deficit
                    excess_reserved[largest.inverter_ids] = 0.0
            if deficit < -0.1:
                left_over = power_w - distributed_power
                if left_over > -deficit:
                    distributed_power += deficit
                elif left_over > 0.0:
                    distributed_power += left_over

        for inverter_ids, excess in excess_reserved.items():
            distributed_power += excess
            battery_power = distribution[inverter_ids]
            battery_power.power += excess
            # Add excess power to the inverter set
            distribution[inverter_ids] = battery_power

        left_over = power_w - distributed_power

        distribution, left_over = self._greedy_distribute_remaining_power(
            distribution, left_over
        )
        inverter_distribution = self._distribute_multi_inverter_pairs(
            distribution, excl_bounds, incl_bounds
        )

        return DistributionResult(
            distribution=inverter_distribution, remaining_power=left_over
        )

    def _distribute_multi_inverter_pairs(
        self,
        distribution: dict[_InverterSet, _Power],
        excl_bounds: dict[int, float],
        incl_bounds: dict[int, float],
    ) -> dict[int, float]:
        """Distribute power between inverters in a set for a single pair.

        Args:
            distribution: distribution with key: inverter_ids, value: (battery_id, power)
            excl_bounds: exclusion bounds for inverters and batteries
            incl_bounds: inclusion bounds for inverters and batteries

        Returns:
            Return the power for each inverter in given distribution.
        """
        new_distribution: dict[int, float] = {}

        for inverter_ids, power in distribution.items():
            if len(inverter_ids) == 1:
                inverter_id = next(iter(inverter_ids))
                new_distribution[inverter_id] = power.power
            else:
                remaining_power = power.power

                # Inverters are sorted by largest excl bound first
                for inverter_id in inverter_ids:
                    if (
                        not is_close_to_zero(remaining_power)
                        and excl_bounds[inverter_id] <= remaining_power
                    ):
                        new_power = min(incl_bounds[inverter_id], remaining_power)

                        new_distribution[inverter_id] = new_power
                        remaining_power -= new_power
                    else:
                        new_distribution[inverter_id] = 0.0

        return new_distribution

    def _greedy_distribute_remaining_power(
        self, distribution: dict[_InverterSet, _Power], remaining_power: float
    ) -> tuple[dict[_InverterSet, _Power], float]:
        """Add remaining power greedily to the given distribution.

        Distribution for each inverter will not exceed its upper bound.

        Args:
            distribution: distribution
            remaining_power: power to distribute

        Returns:
            Return the new distribution and remaining power.
        """
        if is_close_to_zero(remaining_power):
            return distribution, remaining_power

        for inverter_ids, power in distribution.items():
            # The power.power == 0 means the inverter shall not be used due to
            # SoC bounds or no capacity
            if is_close_to_zero(remaining_power) or is_close_to_zero(power.power):
                distribution[inverter_ids] = power
            else:
                additional_power = min(power.upper_bound - power.power, remaining_power)
                power.power += additional_power
                remaining_power -= additional_power

        return distribution, remaining_power

    def distribute_power_equally(
        self, power: float, inverters: set[int]
    ) -> DistributionResult:
        """Distribute the power equally between the inverters in the set.

        This function is mainly useful to set the power for components that are
        broken or have no metrics available.

        Args:
            power: the power to distribute.
            inverters: the inverters to set the power to.

        Returns:
            the power distribution result.
        """
        power_per_inverter = power / len(inverters)
        return DistributionResult(
            distribution={id: power_per_inverter for id in inverters},
            remaining_power=0.0,
        )

    def distribute_power(
        self, power: float, components: list[InvBatPair]
    ) -> DistributionResult:
        """Distribute given power between given components.

        Args:
            power: Power to distribute
            components: InvBatPaired components data. Each pair should have data
                for battery and adjacent inverter.

        Returns:
            Distribution result
        """
        if is_close_to_zero(power):
            return DistributionResult(
                distribution={
                    inverter.component_id: 0.0
                    for _, inverters in components
                    for inverter in inverters
                },
                remaining_power=0.0,
            )
        if power > 0.0:
            return self._distribute_consume_power(power, components)
        return self._distribute_supply_power(power, components)

    def _distribute_consume_power(
        self, power_w: float, components: list[InvBatPair]
    ) -> DistributionResult:
        """Distribute power between the given components.

        Distribute power in a way that the SoC level between given components will:
            * stay on the same level, equal in all given components
            * will try to align himself to the same level.

        Args:
            power_w: power to distribute
            components: list of components between which the power should be
                distributed.

        Returns:
            Distribution result, batteries with no SoC and capacity won't be used.
        """
        # If SoC exceeded bound then remaining SoC should be 0.
        # Otherwise algorithm would try to supply power from that battery
        # in order to keep equal SoC level.
        available_soc: dict[int, float] = {}
        for battery, _ in components:
            available_soc[battery.component_id] = max(
                0.0, battery.soc_upper_bound - battery.soc
            )

        incl_bounds, excl_bounds = self._inclusion_exclusion_bounds(
            components, supply=False
        )

        return self._distribute_power(
            components=components,
            power_w=power_w,
            available_soc=available_soc,
            incl_bounds=incl_bounds,
            excl_bounds=excl_bounds,
        )

    def _distribute_supply_power(
        self, power_w: float, components: list[InvBatPair]
    ) -> DistributionResult:
        """Distribute power between the given components.

        Distribute power in a way that the SoC level between given components will:
            * stay on the same level, equal in all given components
            * will try to align himself to the same level.

        Args:
            power_w: power to distribute
            components: list of components between which the power should be
                distributed.

        Returns:
            Distribution result.
        """
        available_soc: dict[int, float] = {}
        for battery, _ in components:
            available_soc[battery.component_id] = max(
                0.0, battery.soc - battery.soc_lower_bound
            )

        incl_bounds, excl_bounds = self._inclusion_exclusion_bounds(
            components, supply=True
        )

        result: DistributionResult = self._distribute_power(
            components=components,
            power_w=-1 * power_w,
            available_soc=available_soc,
            incl_bounds=incl_bounds,
            excl_bounds=excl_bounds,
        )

        for inverter_id in result.distribution.keys():
            result.distribution[inverter_id] *= -1
        result.remaining_power *= -1

        return result

    def _inclusion_exclusion_bounds(
        self, components: list[InvBatPair], supply: bool = False
    ) -> tuple[dict[int, float], dict[int, float]]:
        """Calculate inclusion and exclusion bounds for given components.

        Inverter exclusion bounds are _not_ adjusted to battery inclusion
        bounds, as the battery exclusion bounds can be satisfied by multiple
        inverters with lower exclusion bounds.

        Args:
            components: list of components.
            supply: if True then supply bounds will be calculated, otherwise
                consume bounds.

        Returns:
            inclusion and exclusion bounds.
        """
        incl_bounds: dict[int, float] = {}
        excl_bounds: dict[int, float] = {}
        for battery, inverters in components:
            if supply:
                excl_bounds[battery.component_id] = (
                    -battery.power_bounds.exclusion_lower
                )
                incl_bounds[battery.component_id] = (
                    -battery.power_bounds.inclusion_lower
                )
            else:
                excl_bounds[battery.component_id] = battery.power_bounds.exclusion_upper
                incl_bounds[battery.component_id] = battery.power_bounds.inclusion_upper

            for inverter in inverters:
                if supply:
                    incl_bounds[inverter.component_id] = -max(
                        inverter.active_power_inclusion_lower_bound,
                        battery.power_bounds.inclusion_lower,
                    )
                    excl_bounds[inverter.component_id] = (
                        -inverter.active_power_exclusion_lower_bound
                    )

                else:
                    incl_bounds[inverter.component_id] = min(
                        inverter.active_power_inclusion_upper_bound,
                        battery.power_bounds.inclusion_upper,
                    )
                    excl_bounds[inverter.component_id] = (
                        inverter.active_power_exclusion_upper_bound
                    )
        return incl_bounds, excl_bounds



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Manage batteries and inverters for the power distributor."""

import abc
import collections.abc

from frequenz.channels import Sender

from .._component_status import ComponentPoolStatus
from ..request import Request
from ..result import Result


class ComponentManager(abc.ABC):
    """Abstract class to manage the data streams for components."""

    @abc.abstractmethod
    def __init__(
        self,
        component_pool_status_sender: Sender[ComponentPoolStatus],
        results_sender: Sender[Result],
    ):
        """Initialize the component data manager.

        Args:
            component_pool_status_sender: Channel for sending information about which
                components are expected to be working.
            results_sender: Channel for sending the results of power distribution.
        """

    @abc.abstractmethod
    def component_ids(self) -> collections.abc.Set[int]:
        """Return the set of component ids."""

    @abc.abstractmethod
    async def start(self) -> None:
        """Start the component data manager."""

    @abc.abstractmethod
    async def distribute_power(self, request: Request) -> None:
        """Distribute the requested power to the components.

        Args:
            request: Request to get the distribution for.

        Returns:
            Result of the distribution.
        """

    @abc.abstractmethod
    async def stop(self) -> None:
        """Stop the component data manager."""



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Manage EV chargers for the power distributor."""

import asyncio
import collections.abc
import logging
from datetime import datetime, timedelta, timezone

from frequenz.channels import (
    Broadcast,
    LatestValueCache,
    Sender,
    merge,
    select,
    selected_from,
)
from frequenz.client.microgrid import (
    ApiClientError,
    ComponentCategory,
    EVChargerData,
    MicrogridApiClient,
)
from frequenz.quantities import Power, Voltage
from typing_extensions import override

from ....._internal._asyncio import run_forever
from ....._internal._math import is_close_to_zero
from .....timeseries import Sample3Phase
from .... import _data_pipeline, connection_manager
from ..._component_pool_status_tracker import ComponentPoolStatusTracker
from ..._component_status import ComponentPoolStatus, EVChargerStatusTracker
from ...request import Request
from ...result import PartialFailure, Result, Success
from .._component_manager import ComponentManager
from ._config import EVDistributionConfig
from ._states import EvcState, EvcStates

_logger = logging.getLogger(__name__)


class EVChargerManager(ComponentManager):
    """Manage ev chargers for the power distributor."""

    @override
    def __init__(
        self,
        component_pool_status_sender: Sender[ComponentPoolStatus],
        results_sender: Sender[Result],
        api_power_request_timeout: timedelta,
    ):
        """Initialize the ev charger data manager.

        Args:
            component_pool_status_sender: Channel for sending information about which
                components are expected to be working.
            results_sender: Channel for sending results of power distribution.
            api_power_request_timeout: Timeout to use when making power requests to
                the microgrid API.
        """
        self._results_sender = results_sender
        self._api_power_request_timeout = api_power_request_timeout
        self._ev_charger_ids = self._get_ev_charger_ids()
        self._evc_states = EvcStates()
        self._voltage_cache: LatestValueCache[Sample3Phase[Voltage]] = LatestValueCache(
            _data_pipeline.voltage_per_phase().new_receiver(),
            unique_id=f"{type(self).__name__}«{hex(id(self))}»:voltage_cache",
        )
        self._config = EVDistributionConfig(component_ids=self._ev_charger_ids)
        self._component_pool_status_tracker = ComponentPoolStatusTracker(
            component_ids=self._ev_charger_ids,
            component_status_sender=component_pool_status_sender,
            max_data_age=timedelta(seconds=10.0),
            max_blocking_duration=timedelta(seconds=30.0),
            component_status_tracker_type=EVChargerStatusTracker,
        )
        self._target_power = Power.zero()
        self._target_power_channel = Broadcast[Request](name="target_power")
        self._target_power_tx = self._target_power_channel.new_sender()
        self._task: asyncio.Task[None] | None = None
        self._latest_request: Request = Request(Power.zero(), set())

    @override
    def component_ids(self) -> collections.abc.Set[int]:
        """Return the set of ev charger ids."""
        return self._ev_charger_ids

    @override
    async def start(self) -> None:
        """Start the ev charger data manager."""
        # Need to start a task only if there are EV chargers in the component graph.
        if self._ev_charger_ids:
            self._task = asyncio.create_task(run_forever(self._run))

    @override
    async def distribute_power(self, request: Request) -> None:
        """Distribute the requested power to the ev chargers.

        Args:
            request: Request to get the distribution for.
        """
        if self._ev_charger_ids:
            await self._target_power_tx.send(request)

    @override
    async def stop(self) -> None:
        """Stop the ev charger manager."""
        await self._voltage_cache.stop()
        await self._component_pool_status_tracker.stop()

    def _get_ev_charger_ids(self) -> collections.abc.Set[int]:
        """Return the IDs of all EV chargers present in the component graph."""
        return {
            evc.component_id
            for evc in connection_manager.get().component_graph.components(
                component_categories={ComponentCategory.EV_CHARGER}
            )
        }

    def _allocate_new_ev(self, component_id: int) -> dict[int, Power]:
        """Allocate power to a newly connected EV charger.

        Args:
            component_id: ID of the EV charger to allocate power to.

        Returns:
            A dictionary containing updated power allocations for the EV chargers.
        """
        available_power = (
            self._target_power - self._evc_states.get_total_allocated_power()
        )
        voltage = self._voltage_cache.get().min()
        if voltage is None:
            _logger.warning(
                "Voltage data is not available. Cannot allocate power to EV charger %s",
                component_id,
            )
            return {}
        initial_power = voltage * self._config.initial_current * 3.0
        if available_power > initial_power:
            return {component_id: initial_power}

        min_power = voltage * self._config.min_current * 3.0
        if available_power > min_power:
            return {component_id: min_power}

        return {}

    def _act_on_new_data(self, ev_data: EVChargerData) -> dict[int, Power]:
        """Act on new data from an EV charger.

        Args:
            ev_data: New data from the EV charger.

        Returns:
            A dictionary containing updated power allocations for the EV chargers.
        """
        component_id = ev_data.component_id
        ev_connected = ev_data.is_ev_connected()
        ev_previously_connected = self._evc_states.get(
            component_id
        ).last_data.is_ev_connected()

        # if EV is just connected, try to set config.initial_current, throttle other
        # EVs if necessary
        ev_newly_connected = ev_connected and not ev_previously_connected
        if ev_newly_connected:
            _logger.info("New EV connected to EV charger %s", component_id)
            return self._allocate_new_ev(component_id)

        # if EV is disconnected, set limit to 0.0.  redistribution to other EVs will
        # happen separately, when possible.
        if not ev_connected:
            if ev_previously_connected:
                _logger.info("EV disconnected from EV charger %s", component_id)
            if self._evc_states.get(component_id).last_allocation > Power.zero():
                return {component_id: Power.zero()}

        # else if last throttling was less than 'increase_power_interval', do nothing.
        now = datetime.now(tz=timezone.utc)
        last_throttling_time = self._evc_states.get(component_id).last_reallocation_time
        if last_throttling_time is not None:
            dur = now - last_throttling_time
            if dur < self._config.increase_power_interval:
                return {}

        # if ev's target power was previously set to zero, treat it like it is newly
        # connected
        evc = self._evc_states.get(component_id)
        if is_close_to_zero(evc.last_allocation.as_watts()):
            return self._allocate_new_ev(component_id)

        # if the ev charger is already allocated the max power, do nothing
        allottable_power = Power.from_watts(
            evc.last_data.active_power_inclusion_upper_bound
            - evc.last_allocation.as_watts()
        )
        available_power = (
            self._target_power - self._evc_states.get_total_allocated_power()
        )
        allottable_power = min(allottable_power, available_power)

        if (
            is_close_to_zero(allottable_power.as_watts())
            or allottable_power < Power.zero()
        ):
            return {}

        target_power = min(
            evc.last_allocation + allottable_power,
            Power.from_watts(evc.last_data.active_power_inclusion_upper_bound),
        )
        _logger.debug(
            "Increasing power to EV charger %s from %s to %s",
            component_id,
            evc.last_allocation,
            target_power,
        )
        return {component_id: target_power}

    async def _run(self) -> None:  # pylint: disable=too-many-locals
        """Run the main event loop of the EV charger manager."""
        api = connection_manager.get().api_client
        ev_charger_data_rx = merge(
            *[await api.ev_charger_data(evc_id) for evc_id in self._ev_charger_ids]
        )
        target_power_rx = self._target_power_channel.new_receiver()
        latest_target_powers: dict[int, Power] = {}
        async for selected in select(ev_charger_data_rx, target_power_rx):
            target_power_changes = {}
            now = datetime.now(tz=timezone.utc)

            if selected_from(selected, ev_charger_data_rx):
                evc_data = selected.message
                # If a new ev charger is added, add it to the state tracker, with
                # now as the last reallocation time and last charging time.
                #
                # This means it won't be assigned any power until the reallocation
                # duration has passed.
                if evc_data.component_id not in self._evc_states:
                    self._evc_states.add_evc(
                        EvcState(
                            component_id=evc_data.component_id,
                            last_data=evc_data,
                            power=Power.zero(),
                            last_allocation=Power.zero(),
                            last_reallocation_time=now,
                            last_charging_time=now,
                        )
                    )
                    target_power_changes = {evc_data.component_id: Power.zero()}

                # See if the ev charger has room for more power, and if the last
                # allocation was not in the last reallocation duration.
                else:
                    target_power_changes = self._act_on_new_data(evc_data)
                    self._evc_states.get(evc_data.component_id).update_state(evc_data)

            elif selected_from(selected, target_power_rx):
                self._latest_request = selected.message
                self._target_power = selected.message.power
                _logger.debug("New target power: %s", self._target_power)
                used_power = self._evc_states.get_ev_total_used_power()
                allocated_power = self._evc_states.get_total_allocated_power()
                if self._target_power < used_power:
                    diff_power = used_power - self._target_power
                    target_power_changes = self._throttle_ev_chargers(diff_power)
                    _logger.warning(
                        "Throttling EV chargers by %s-%s=%s: %s",
                        used_power,
                        self._target_power,
                        diff_power,
                        target_power_changes,
                    )
                elif self._target_power < allocated_power:
                    diff_power = allocated_power - self._target_power
                    target_power_changes = self._deallocate_unused_power(diff_power)

            if target_power_changes:
                _logger.debug("Setting power to EV chargers: %s", target_power_changes)
            else:
                continue
            for component_id, power in target_power_changes.items():
                self._evc_states.get(component_id).update_last_allocation(power, now)

            latest_target_powers.update(target_power_changes)
            result = await self._set_api_power(
                api, target_power_changes, self._api_power_request_timeout
            )
            await self._results_sender.send(result)

    async def _set_api_power(
        self,
        api: MicrogridApiClient,
        target_power_changes: dict[int, Power],
        api_request_timeout: timedelta,
    ) -> Result:
        """Send the EV charger power changes to the microgrid API.

        Args:
            api: The microgrid API client to use for setting the power.
            target_power_changes: A dictionary containing the new power allocations for
                the EV chargers.
            api_request_timeout: The timeout for the API request.

        Returns:
            Power distribution result, corresponding to the result of the API
                request.
        """
        tasks: dict[int, asyncio.Task[None]] = {}
        for component_id, power in target_power_changes.items():
            tasks[component_id] = asyncio.create_task(
                api.set_power(component_id, power.as_watts())
            )
        _, pending = await asyncio.wait(
            tasks.values(),
            timeout=api_request_timeout.total_seconds(),
            return_when=asyncio.ALL_COMPLETED,
        )
        for task in pending:
            task.cancel()
        await asyncio.gather(*pending, return_exceptions=True)

        failed_components: set[int] = set()
        succeeded_components: set[int] = set()
        failed_power = Power.zero()
        for component_id, task in tasks.items():
            try:
                task.result()
            except asyncio.CancelledError:
                _logger.warning(
                    "Timeout while setting power to EV charger %s", component_id
                )
            except ApiClientError as exc:
                _logger.warning(
                    "Got a client error while setting power to EV charger %s: %s",
                    component_id,
                    exc,
                )
            except Exception:  # pylint: disable=broad-except
                _logger.exception(
                    "Unknown error while setting power to EV charger: %s", component_id
                )
            else:
                succeeded_components.add(component_id)
                continue

            failed_components.add(component_id)
            failed_power += target_power_changes[component_id]

        if failed_components:
            return PartialFailure(
                failed_components=failed_components,
                succeeded_components=succeeded_components,
                failed_power=failed_power,
                succeeded_power=self._target_power - failed_power,
                excess_power=Power.zero(),
                request=self._latest_request,
            )
        return Success(
            succeeded_components=succeeded_components,
            succeeded_power=self._target_power,
            excess_power=Power.zero(),
            request=self._latest_request,
        )

    def _deallocate_unused_power(self, to_deallocate: Power) -> dict[int, Power]:
        """Reduce the power allocated to the EV chargers to meet the target power.

        This prioritizes reducing power to EV chargers that aren't consuming the
        allocated power.

        Args:
            to_deallocate: The amount of power to reduce the total allocated power by.

        Returns:
            A list of new (reduced) charging current limits for a subset of ev
                chargers, required to bring down the consumption by the given value.
        """
        voltage = self._voltage_cache.get().min()
        if voltage is None:
            _logger.warning(
                "Voltage data is unavailable. Can't deallocate power from EV chargers",
            )
            return {}
        min_power = voltage * self._config.min_current * 3.0

        evc_list = list(self._evc_states.values())
        evc_list.sort(key=lambda st: st.last_allocation - st.power, reverse=True)

        deallocated_power = Power.zero()
        target_power_changes = {}

        for evc in evc_list:
            if deallocated_power >= to_deallocate:
                break
            ev_to_deallocate = evc.last_allocation - evc.power
            if ev_to_deallocate <= Power.zero():
                continue
            ev_to_deallocate = min(ev_to_deallocate, to_deallocate - deallocated_power)
            tgt_power = evc.last_allocation - ev_to_deallocate
            if tgt_power < min_power:
                tgt_power = Power.zero()
            deallocated_power += evc.last_allocation - tgt_power
            target_power_changes[evc.component_id] = tgt_power
        return target_power_changes

    def _throttle_ev_chargers(  # pylint: disable=too-many-locals
        self,
        throttle_by: Power,
    ) -> dict[int, Power]:
        """Reduce EV charging power to meet the target power.

        This targets EV chargers that are currently consuming the most.

        Level 1 throttling is done by reducing the power to the minimum current required
        to charge the EV.  When the consumption is still above the target power, level 2
        throttling is done by reducing the power to 0.

        Args:
            throttle_by: The amount of power to reduce the total EV charging power by.

        Returns:
            A list of new (reduced) charging current limits for a subset of ev
                chargers, required to bring down the consumption by the given value.
        """
        min_power = Power.zero()
        voltage = self._voltage_cache.get().min()
        if voltage is None:
            _logger.warning(
                "Voltage data is not available. Cannot perform level 1 throttling.",
            )
        else:
            min_power = voltage * self._config.min_current * 3.0

        evc_list = list(self._evc_states.values())
        evc_list.sort(key=lambda st: (st.power, st.last_allocation), reverse=True)

        level1_throttling_count = 0
        level1_amps_achieved = Power.zero()

        level2_throttling_count = 0
        level2_amps_achieved = Power.zero()

        for evc in evc_list:
            evc_power = evc.power
            evc_level1_power = Power.zero()
            if evc_power > min_power:
                evc_level1_power = evc_power - min_power

            if evc_power == Power.zero():
                evc_power = evc.last_allocation

            if evc_power == Power.zero():
                break

            if level1_amps_achieved < throttle_by:
                level1_amps_achieved += evc_level1_power
                level1_throttling_count += 1
            else:
                break
            if level2_amps_achieved < throttle_by:
                level2_amps_achieved += evc_power
                level2_throttling_count += 1

        if level1_amps_achieved >= throttle_by:
            throttled_powers = {
                evc.component_id: min_power
                for evc in evc_list[:level1_throttling_count]
            }
        else:
            throttled_powers = {
                evc.component_id: Power.zero()
                for evc in evc_list[:level2_throttling_count]
            }
        _logger.debug("Throttling: %s", throttled_powers)
        return throttled_powers



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Timeseries basic types."""

import dataclasses
import functools
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Generic, Protocol, Self, TypeVar, cast, overload

from frequenz.quantities import Power, Quantity

UNIX_EPOCH = datetime.fromtimestamp(0.0, tz=timezone.utc)
"""The UNIX epoch (in UTC)."""

QuantityT = TypeVar("QuantityT", bound=Quantity)
"""Type variable for representing various quantity types."""


@dataclass(frozen=True, order=True)
class Sample(Generic[QuantityT]):
    """A measurement taken at a particular point in time.

    The `value` could be `None` if a component is malfunctioning or data is
    lacking for another reason, but a sample still needs to be sent to have a
    coherent view on a group of component metrics for a particular timestamp.
    """

    timestamp: datetime
    """The time when this sample was generated."""

    value: QuantityT | None = None
    """The value of this sample."""

    def __str__(self) -> str:
        """Return a string representation of the sample."""
        return f"{type(self).__name__}({self.timestamp}, {self.value})"

    def __repr__(self) -> str:
        """Return a string representation of the sample."""
        return f"{type(self).__name__}({self.timestamp=}, {self.value=})"


@dataclass(frozen=True)
class Sample3Phase(Generic[QuantityT]):
    """A 3-phase measurement made at a particular point in time.

    Each of the `value` fields could be `None` if a component is malfunctioning
    or data is lacking for another reason, but a sample still needs to be sent
    to have a coherent view on a group of component metrics for a particular
    timestamp.
    """

    timestamp: datetime
    """The time when this sample was generated."""
    value_p1: QuantityT | None
    """The value of the 1st phase in this sample."""

    value_p2: QuantityT | None
    """The value of the 2nd phase in this sample."""

    value_p3: QuantityT | None
    """The value of the 3rd phase in this sample."""

    def __iter__(self) -> Iterator[QuantityT | None]:
        """Return an iterator that yields values from each of the phases.

        Yields:
            Per-phase measurements one-by-one.
        """
        yield self.value_p1
        yield self.value_p2
        yield self.value_p3

    @overload
    def max(self, default: QuantityT) -> QuantityT: ...

    @overload
    def max(self, default: None = None) -> QuantityT | None: ...

    def max(self, default: QuantityT | None = None) -> QuantityT | None:
        """Return the max value among all phases, or default if they are all `None`.

        Args:
            default: value to return if all phases are `None`.

        Returns:
            Max value among all phases, if available, default value otherwise.
        """
        if not any(self):
            return default
        value: QuantityT = functools.reduce(
            lambda x, y: x if x > y else y,
            filter(None, self),
        )
        return value

    @overload
    def min(self, default: QuantityT) -> QuantityT: ...

    @overload
    def min(self, default: None = None) -> QuantityT | None: ...

    def min(self, default: QuantityT | None = None) -> QuantityT | None:
        """Return the min value among all phases, or default if they are all `None`.

        Args:
            default: value to return if all phases are `None`.

        Returns:
            Min value among all phases, if available, default value otherwise.
        """
        if not any(self):
            return default
        value: QuantityT = functools.reduce(
            lambda x, y: x if x < y else y,
            filter(None, self),
        )
        return value

    def map(
        self,
        function: Callable[[QuantityT], QuantityT],
        default: QuantityT | None = None,
    ) -> Self:
        """Apply the given function on each of the phase values and return the result.

        If a phase value is `None`, replace it with `default` instead.

        Args:
            function: The function to apply on each of the phase values.
            default: The value to apply if a phase value is `None`.

        Returns:
            A new instance, with the given function applied on values for each of the
                phases.
        """
        return self.__class__(
            timestamp=self.timestamp,
            value_p1=default if self.value_p1 is None else function(self.value_p1),
            value_p2=default if self.value_p2 is None else function(self.value_p2),
            value_p3=default if self.value_p3 is None else function(self.value_p3),
        )


class Comparable(Protocol):
    """A protocol that requires the implementation of comparison methods.

    This protocol is used to ensure that types can be compared using
    the less than or equal to (`<=`) and greater than or equal to (`>=`)
    operators.
    """

    def __le__(self, other: Any, /) -> bool:
        """Return whether this instance is less than or equal to `other`."""

    def __ge__(self, other: Any, /) -> bool:
        """Return whether this instance is greater than or equal to `other`."""


_T = TypeVar("_T", bound=Comparable | None)


@dataclass(frozen=True)
class Bounds(Generic[_T]):
    """Lower and upper bound values."""

    lower: _T
    """Lower bound."""

    upper: _T
    """Upper bound."""

    def __contains__(self, item: _T) -> bool:
        """
        Check if the value is within the range of the container.

        Args:
            item: The value to check.

        Returns:
            bool: True if value is within the range, otherwise False.
        """
        if self.lower is None and self.upper is None:
            return True
        if self.lower is None:
            return item <= self.upper
        if self.upper is None:
            return self.lower <= item

        return cast(Comparable, self.lower) <= item <= cast(Comparable, self.upper)


@dataclass(frozen=True, kw_only=True)
class SystemBounds:
    """Internal representation of system bounds for groups of components."""

    # compare = False tells the dataclass to not use name for comparison methods
    timestamp: datetime = dataclasses.field(compare=False)
    """Timestamp of the metrics."""

    inclusion_bounds: Bounds[Power] | None
    """Total inclusion power bounds for all components of a pool.

    This is the range within which power requests would be allowed by the pool.

    When exclusion bounds are present, they will exclude a subset of the inclusion
    bounds.
    """

    exclusion_bounds: Bounds[Power] | None
    """Total exclusion power bounds for all components of a pool.

    This is the range within which power requests are NOT allowed by the pool.
    If present, they will be a subset of the inclusion bounds.
    """

    def __contains__(self, item: Power) -> bool:
        """
        Check if the value is within the range of the container.

        Args:
            item: The value to check.

        Returns:
            bool: True if value is within the range, otherwise False.
        """
        if not self.inclusion_bounds or item not in self.inclusion_bounds:
            return False
        if self.exclusion_bounds and item in self.exclusion_bounds:
            return False
        return True



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Background service that tracks the status of an EV charger."""


import asyncio
import logging
from datetime import datetime, timedelta, timezone

from frequenz.channels import Receiver, Sender, select, selected_from
from frequenz.channels.timer import SkipMissedAndDrift, Timer
from frequenz.client.microgrid import (
    EVChargerCableState,
    EVChargerComponentState,
    EVChargerData,
)
from typing_extensions import override

from ...._internal._asyncio import run_forever
from ....actor._background_service import BackgroundService
from ... import connection_manager
from ._blocking_status import BlockingStatus
from ._component_status import (
    ComponentStatus,
    ComponentStatusEnum,
    ComponentStatusTracker,
    SetPowerResult,
)

_logger = logging.getLogger(__name__)


class EVChargerStatusTracker(ComponentStatusTracker, BackgroundService):
    """Status tracker for EV chargers.

    It reports an EV charger as `WORKING` when an EV is connected to it,
    and power can be allocated to it, and `NOT_WORKING` otherwise.

    If it receives a power assignment failure from the PowerDistributor,
    when the component is expected to be `WORKING`, it is marked as
    `UNCERTAIN` for a specific interval, before being marked `WORKING`
    again.
    """

    @override
    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        component_id: int,
        max_data_age: timedelta,
        max_blocking_duration: timedelta,
        status_sender: Sender[ComponentStatus],
        set_power_result_receiver: Receiver[SetPowerResult],
    ) -> None:
        """Initialize this instance.

        Args:
            component_id: ID of the EV charger to monitor the status of.
            max_data_age: max duration to wait for, before marking a component as
                NOT_WORKING, unless new data arrives.
            max_blocking_duration: duration for which the component status should be
                UNCERTAIN if a request to the component failed unexpectedly.
            status_sender: Channel sender to send status updates to.
            set_power_result_receiver: Receiver to fetch PowerDistributor responses
                from, to get the status of the most recent request made for an EV
                Charger.
        """
        self._component_id = component_id
        self._max_data_age = max_data_age
        self._status_sender = status_sender
        self._set_power_result_receiver = set_power_result_receiver

        self._last_status = ComponentStatusEnum.NOT_WORKING
        self._blocking_status = BlockingStatus(
            min_duration=timedelta(seconds=1.0),
            max_duration=max_blocking_duration,
        )

        BackgroundService.__init__(self, name=f"EVChargerStatusTracker({component_id})")

    @override
    def start(self) -> None:
        """Start the status tracker."""
        self._tasks.add(asyncio.create_task(run_forever(self._run)))

    def _is_working(self, ev_data: EVChargerData) -> bool:
        """Return whether the given data indicates that the component is working."""
        return ev_data.cable_state in (
            EVChargerCableState.EV_PLUGGED,
            EVChargerCableState.EV_LOCKED,
        ) and ev_data.component_state in (
            EVChargerComponentState.READY,
            EVChargerComponentState.CHARGING,
            EVChargerComponentState.DISCHARGING,
        )

    def _is_stale(self, ev_data: EVChargerData) -> bool:
        """Return whether the given data is stale."""
        now = datetime.now(tz=timezone.utc)
        stale = now - ev_data.timestamp > self._max_data_age
        return stale

    def _handle_ev_data(self, ev_data: EVChargerData) -> ComponentStatusEnum:
        """Handle new EV charger data."""
        if self._is_stale(ev_data):
            if self._last_status == ComponentStatusEnum.WORKING:
                _logger.warning(
                    "EV charger %s data is stale. Last timestamp: %s",
                    self._component_id,
                    ev_data.timestamp,
                )
            return ComponentStatusEnum.NOT_WORKING

        if self._is_working(ev_data):
            if self._last_status == ComponentStatusEnum.NOT_WORKING:
                _logger.warning(
                    "EV charger %s is in WORKING state.",
                    self._component_id,
                )
            return ComponentStatusEnum.WORKING

        if self._last_status == ComponentStatusEnum.WORKING:
            _logger.warning(
                "EV charger %s is in NOT_WORKING state. "
                "Cable state: %s, component state: %s",
                self._component_id,
                ev_data.cable_state,
                ev_data.component_state,
            )
        return ComponentStatusEnum.NOT_WORKING

    def _handle_set_power_result(
        self, set_power_result: SetPowerResult
    ) -> ComponentStatusEnum:
        """Handle a new set power result."""
        if self._component_id in set_power_result.succeeded:
            return ComponentStatusEnum.WORKING

        self._blocking_status.block()
        if self._last_status == ComponentStatusEnum.WORKING:
            _logger.warning(
                "EV charger %s is in UNCERTAIN state. Set power result: %s",
                self._component_id,
                set_power_result,
            )
        return ComponentStatusEnum.UNCERTAIN

    async def _run(self) -> None:
        """Run the status tracker."""
        api_client = connection_manager.get().api_client
        ev_data_rx = await api_client.ev_charger_data(self._component_id)
        set_power_result_rx = self._set_power_result_receiver
        missing_data_timer = Timer(self._max_data_age, SkipMissedAndDrift())

        # Send initial status
        await self._status_sender.send(
            ComponentStatus(self._component_id, self._last_status)
        )

        async for selected in select(
            ev_data_rx, set_power_result_rx, missing_data_timer
        ):
            new_status = ComponentStatusEnum.NOT_WORKING
            if selected_from(selected, ev_data_rx):
                missing_data_timer.reset()
                new_status = self._handle_ev_data(selected.message)
            elif selected_from(selected, set_power_result_rx):
                new_status = self._handle_set_power_result(selected.message)
            elif selected_from(selected, missing_data_timer):
                _logger.warning(
                    "No EV charger %s data received for %s. Setting status to NOT_WORKING.",
                    self._component_id,
                    self._max_data_age,
                )

            # Send status update if status changed
            if (
                self._blocking_status.is_blocked()
                and new_status != ComponentStatusEnum.NOT_WORKING
            ):
                new_status = ComponentStatusEnum.UNCERTAIN

            if new_status != self._last_status:
                _logger.info(
                    "EV charger %s status changed from %s to %s",
                    self._component_id,
                    self._last_status,
                    new_status,
                )
                self._last_status = new_status
                await self._status_sender.send(
                    ComponentStatus(self._component_id, new_status)
                )



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

# Please note this docstring is included in the getting started documentation in docs/.
"""Actors are a primitive unit of computation that runs autonomously.

## Actor Programming Model

!!! quote "From [Wikipedia](https://en.wikipedia.org/wiki/Actor_model)"

    The actor model in computer science is a mathematical model of concurrent
    computation that treats an actor as the basic building block of concurrent
    computation. In response to a message it receives, an actor can: make local
    decisions, create more actors, send more messages, and determine how to
    respond to the next message received. Actors may modify their own private
    state, but can only affect each other indirectly through messaging (removing
    the need for lock-based synchronization).

We won't get into much more detail here because it is outside the scope of this
documentation. However, if you are interested in learning more about the actor
programming model, here are some useful resources:

- [Actor Model (Wikipedia)](https://en.wikipedia.org/wiki/Actor_model)
- [How the Actor Model Meets the Needs of Modern, Distributed Systems
  (Akka)](https://doc.akka.io/docs/akka/current/typed/guide/actors-intro.html)

## Frequenz SDK Actors

The [`Actor`][frequenz.sdk.actor.Actor] class serves as the foundation for
creating concurrent tasks and all actors in the SDK inherit from it. This class
provides a straightforward way to implement actors. It shares similarities with
the traditional actor programming model but also has some unique features:

- **Message Passing:** Like traditional actors, our [`Actor`][frequenz.sdk.actor.Actor]
  class communicates through message passing. Even when no particular message passing
  mechanism is enforced, the SDK actors use [`frequenz.channels`][] for communication.

- **Automatic Restart:** If an unhandled exception occurs in an actor's logic
  ([`_run()`][_run] method), the actor will be automatically restarted. This ensures
  robustness in the face of errors.

- **Simplified Task Management:** Actors manage asynchronous tasks using
  [`asyncio`][]. You can create and manage tasks within the actor, and the
  [`Actor`][frequenz.sdk.actor.Actor] class handles task cancellation and cleanup.

- **Simplified lifecycle management:** Actors are [async context managers] and also
  a [`run()`][frequenz.sdk.actor.run] function is provided to easily run a group of
  actors and wait for them to finish.

## Lifecycle


Actors are not started when they are created. There are 3 main ways to start an actor
(from most to least recommended):

1. By using the [`run()`][frequenz.sdk.actor.run] function.
2. By using the actor as an [async context manager].
3. By using the [`start()`][frequenz.sdk.actor.Actor.start] method.

!!! warning

    1. If an actor raises an unhandled exception, it will be restarted automatically.

    2. Actors manage [`asyncio.Task`][] objects, so a reference to the actor object must
       be held for as long as the actor is expected to be running, otherwise its tasks
       will be cancelled and the actor will stop. For more information, please refer to
       the [Python `asyncio`
       documentation](https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task).


### The `run()` Function

The [`run()`][frequenz.sdk.actor.run] function can start many actors at once and waits
for them to finish. If any of the actors is stopped with errors, the errors will be
logged.

???+ example

    ```python
    import asyncio
    from frequenz.sdk.actor import Actor, run

    class MyActor(Actor):
        async def _run(self) -> None:
            while True:
                print("Hello World!")
                await asyncio.sleep(1)

    await asyncio.run(MyActor()) # (1)!
    ```

    1. This line will block until the actor completes its execution or is manually stopped.

### Async Context Manager

When an actor is used as an [async context manager], it is started when the
`async with` block is entered and stopped automatically when the block is exited
(even if an unhandled exception is raised).

???+ example

    ```python
    from frequenz.sdk.actor import Actor

    class MyActor(Actor):
        async def _run(self) -> None:
            print("Hello World!")

    async with MyActor() as actor: # (1)!
        print("The actor is running")
    # (2)!
    ```

    1. [`start()`][frequenz.sdk.actor.Actor.start] is called automatically when entering
        the `async with` block.
    2. [`stop()`][frequenz.sdk.actor.Actor.stop] is called automatically when exiting
        the `async with` block.

### The `start()` Method

When using the [`start()`][frequenz.sdk.actor.Actor.start] method, the actor is
started in the background and the method returns immediately. The actor will
continue to run until it is **manually** stopped or until it completes its execution.

???+ example

    ```python
    from frequenz.sdk.actor import Actor

    class MyActor(Actor):
        async def _run(self) -> None:
            print("Hello World!")

    actor = MyActor() # (1)!
    actor.start() # (2)!
    print("The actor is running") # (3)!
    await actor.stop() # (4)!
    ```

    1. The actor is created but not started yet.
    2. The actor is started manually, it keeps running in the background.
    3. !!! danger

        If this function would raise an exception, the actor will never be stopped!

    4. Until the actor is stopped manually.

!!! warning

    This method is not recommended because it is easy to forget to stop the actor
    manually, especially in error conditions.

## Communication

The [`Actor`][frequenz.sdk.actor.Actor] class doesn't impose any specific way to
communicate between actors. However, [`frequenz.channels`][] are always used as the
communication mechanism between actors in the SDK.

## Implementing an Actor

To implement an actor, you must inherit from the [`Actor`][frequenz.sdk.actor.Actor]
class and implement an *initializer* and the abstract [`_run()`][_run] method.

### Initialization

The [initializer][object.__init__] is called when the actor is created. The
[`Actor`][frequenz.sdk.actor.Actor] class initializer
([`__init__`][frequenz.sdk.actor.Actor.__init__]) should be always called first in the
class we are implementing to make sure actors are properly initialized.

The [`Actor.__init__()`][frequenz.sdk.actor.Actor.__init__] takes one optional argument,
a [`name`][frequenz.sdk.actor.Actor.name] that will be used to identify the actor in
logs. If no name is provided, a default name will be generated, but it is recommended
that [`Actor`][frequenz.sdk.actor.Actor] subclasses can also receive a name as an
argument to make it easier to identify individual instances in logs more easily.

The actor initializer normally also accepts as arguments the input channels receivers
and output channels senders that will be used for communication. These channels should
be created outside the actor and passed to it as arguments to ensure actors can be
composed easily.

???+ example "Example echo actor"

    ```python
    from frequenz.channels import Receiver, Sender
    from frequenz.sdk.actor import Actor

    class EchoActor(Actor):  # (1)!
        def __init__(
                self,
                receiver: Receiver[int],  # (2)!
                output: Sender[int],  # (3)!
                name: str | None = None,  # (4)!
        ) -> None:
            super().__init__(name=name) # (5)!
            self._input: Receiver[int] = receiver  # (6)!
            self._output: Sender[int] = output  # (7)!
    ```

    1. We define a new actor class called `EchoActor` that inherits from
        [`Actor`][frequenz.sdk.actor.Actor].

    2. We accept an `receiver` argument that will be used to receive messages from
        a channel.
    3. We accept an `output` argument that will be used to send messages to a channel.
    4. We accept an optional `name` argument that will be used to identify the actor in
        logs.
    5. We call [`Actor.__init__()`][frequenz.sdk.actor.Actor.__init__] to make sure the
        actor is properly initialized.
    6. We store the `receiver` argument in a *private* attribute to use it later.
    7. We store the `output` argument in a *private* attribute to use it later.

### The `_run()` Method

The abstract `_run()` method is called automatically by the base class when the actor is
[started][frequenz.sdk.actor.Actor.start].

Normally an actor should run forever (or until it is
[stopped][frequenz.sdk.actor.Actor.stop]), so it is very common to have an infinite loop
in the `_run()` method, typically receiving messages from one or more channels
([receivers][frequenz.channels.Receiver]), processing them and sending the results to
other channels ([senders][frequenz.channels.Sender]).
However, it is worth noting that an actor can also be designed for a one-time execution
or a limited number of runs, terminating upon completion.

???+ example "Example echo actor"

    ```python
    from frequenz.channels import Receiver, Sender
    from frequenz.sdk.actor import Actor

    class EchoActor(Actor):
        def __init__(
                self,
                receiver: Receiver[int],
                output: Sender[int],
                name: str | None = None,
        ) -> None:
            super().__init__(name=name)
            self._input: Receiver[int] = receiver
            self._output: Sender[int] = output

        async def _run(self) -> None:  # (1)!
            async for msg in self._input:  # (2)!
                await self._output.send(msg)  # (3)!
    ```

    1. We implement the abstract [`_run()`][_run] method.
    2. We receive messages from the `receiver` one by one.
    3. We send the received message to the `output` channel.

### Stopping

By default, the [`stop()`][frequenz.sdk.actor.Actor.stop] method will call the
[`cancel()`][frequenz.sdk.actor.Actor.cancel] method (which will cancel all the tasks
created by the actor) and will wait for them to finish.

This means that when an actor is stopped, the [`_run()`][_run] method will receive
a [`CancelledError`][asyncio.CancelledError] exception. You should have this in mind
when implementing your actor and make sure to handle this exception properly if you need
to do any cleanup.

The actor will handle the [`CancelledError`][asyncio.CancelledError] exception
automatically if it is not handled in the [`_run()`][_run] method, so if there is no
need for extra cleanup, you don't need to worry about it.

If an unhandled exception is raised in the [`_run()`][_run] method, the actor will
re-run the [`_run()`][_run] method automatically. This ensures robustness in the face of
errors, but you should also have this in mind if you need to do any cleanup to make sure
the re-run doesn't cause any problems.

???+ tip

    You could implement your own [`stop()`][frequenz.sdk.actor.Actor.stop] method to,
    for example, send a message to a channel to notify that the actor should stop, or
    any other more graceful way to stop the actor if you need to make sure it can't be
    interrupted at any `await` point.

### Spawning Extra Tasks

Actors run at least one background task, created automatically by the
[`Actor`][frequenz.sdk.actor.Actor] class. But [`Actor`][frequenz.sdk.actor.Actor]
inherits from [`BackgroundService`][frequenz.sdk.actor.BackgroundService], which
provides a few methods to create and manage extra tasks.

If your actor needs to spawn extra tasks, you can use
[`BackgroundService`][frequenz.sdk.actor.BackgroundService] facilities to manage the
tasks, so they are also automatically stopped when the actor is stopped.

All you need to do is add the newly spawned tasks to the actor's
[`tasks`][frequenz.sdk.actor.Actor.tasks] set.

???+ example

    ```python
    import asyncio
    from frequenz.sdk.actor import Actor

    class MyActor(Actor):
        async def _run(self) -> None:
            extra_task = asyncio.create_task(self._extra_task())  # (1)!
            self.tasks.add(extra_task)  # (2)!
            while True:  # (3)!
                print("_run() running")
                await asyncio.sleep(1)

        async def _extra_task(self) -> None:
            while True:  # (4)!
                print("_extra_task() running")
                await asyncio.sleep(1.1)

    async with MyActor() as actor:  # (5)!
        await asyncio.sleep(3)  # (6)!
    # (7)!
    ```

    1. We create a new task using [`asyncio.create_task()`][asyncio.create_task].
    2. We add the task to the actor's [`tasks`][frequenz.sdk.actor.Actor.tasks] set.
        This ensures the task will be cancelled and cleaned up when the actor is stopped.
    3. We leave the actor running forever.
    4. The extra task will also run forever.
    5. The actor is started.
    6. We wait for 3 seconds, the actor should print a bunch of "_run() running" and
        "_extra_task() running" messages while it's running.
    7. The actor is stopped and the extra task is cancelled automatically.

### Examples

Here are a few simple but complete examples to demonstrate how to create actors and
connect them using [channels][frequenz.channels].

!!! tip

    The code examples are annotated with markers (like {{code_annotation_marker}}), you
    can click on them to see the step-by-step explanation of what's going on. The
    annotations are numbered according to the order of execution.

#### Composing actors

This example shows how to create two actors and connect them using
[broadcast][frequenz.channels.Broadcast] channels.

```python title="compose.py"
import asyncio

from frequenz.channels import Broadcast, Receiver, Sender
from frequenz.sdk.actor import Actor

class Actor1(Actor):  # (1)!
    def __init__(
        self,
        receiver: Receiver[str],
        output: Sender[str],
        name: str | None = None,
    ) -> None:
        super().__init__(name=name)
        self._receiver = receiver
        self._output = output

    async def _run(self) -> None:
        async for msg in self._receiver:
            await self._output.send(f"Actor1 forwarding: {msg!r}")  # (8)!


class Actor2(Actor):
    def __init__(
        self,
        receiver: Receiver[str],
        output: Sender[str],
        name: str | None = None,
    ) -> None:
        super().__init__(name=name)
        self._receiver = receiver
        self._output = output

    async def _run(self) -> None:
        async for msg in self._receiver:
            await self._output.send(f"Actor2 forwarding: {msg!r}")  # (9)!


async def main() -> None:  # (2)!
    # (4)!
    input_channel: Broadcast[str] = Broadcast("Input to Actor1")
    middle_channel: Broadcast[str] = Broadcast("Actor1 -> Actor2 stream")
    output_channel: Broadcast[str] = Broadcast("Actor2 output")

    input_sender = input_channel.new_sender()
    output_receiver = output_channel.new_receiver()

    async with (  # (5)!
        Actor1(input_channel.new_receiver(), middle_channel.new_sender(), "actor1"),
        Actor2(middle_channel.new_receiver(), output_channel.new_sender(), "actor2"),
    ):
        await input_sender.send("Hello")  # (6)!
        msg = await output_receiver.receive()  # (7)!
        print(msg)  # (10)!
    # (11)!

if __name__ == "__main__":  # (3)!
    asyncio.run(main())
```

1. We define 2 actors: `Actor1` and `Actor2` that will just forward a message
   from an input channel to an output channel, adding some text.

2. We define an async `main()` function within the main logic of our [asyncio][] program.

3. We start the `main()` function in the async loop using [`asyncio.run()`][asyncio.run].

4. We create a bunch of [broadcast][frequenz.channels.Broadcast]
   [channels][frequenz.channels] to connect our actors.

    * `input_channel` is the input channel for `Actor1`.
    * `middle_channel` is the channel that connects `Actor1` and `Actor2`.
    * `output_channel` is the output channel for `Actor2`.

5. We create two actors and use them as async context managers, `Actor1` and
    `Actor2`, and connect them by creating new
    [senders][frequenz.channels.Sender] and
    [receivers][frequenz.channels.Receiver] from the channels.

    !!! note

        We don't use the [`run()`][frequenz.sdk.actor.run] function here because we
        want to stop the actors when we are done with them, but the actors will run
        forever (as long as the channel is not closed). So the async context manager
        is a better fit for this example.

6. We schedule the [sending][frequenz.channels.Sender.send] of the message
   `Hello` to `Actor1` via `input_channel`.

7. We [receive][frequenz.channels.Receiver.receive] (await) the response from
   `Actor2` via `output_channel`. Between this and the previous steps the
   `async` calls in the actors will be executed.

8. `Actor1` sends the re-formatted message (`Actor1 forwarding: Hello`) to
   `Actor2` via the `middle_channel`.

9. `Actor2` sends the re-formatted message (`Actor2 forwarding: "Actor1
   forwarding: 'Hello'"`) to the `output_channel`.

10. Finally, we print the received message, which will still be `Actor2
    forwarding: "Actor1 forwarding: 'Hello'"`.

11. The actors are stopped and cleaned up automatically when the `async with`
    block ends.

The expected output is:

```
Actor2 forwarding: "Actor1 forwarding: 'Hello'"
```

#### Receiving from multiple channels

This example shows how to create an actor that receives messages from multiple
[broadcast][frequenz.channels.Broadcast] channels using
[`select()`][frequenz.channels.select].

```python title="select.py"
import asyncio

from frequenz.channels import Broadcast, Receiver, Sender
from frequenz.channels.util import select, selected_from
from frequenz.sdk.actor import Actor, run


class EchoActor(Actor):  # (1)!
    def __init__(
        self,
        receiver_1: Receiver[bool],
        receiver_2: Receiver[bool],
        output: Sender[bool],
        name: str | None = None,
    ) -> None:
        super().__init__(name=name)
        self._receiver_1 = receiver_1
        self._receiver_2 = receiver_2
        self._output = output

    async def _run(self) -> None:  # (2)!
        async for selected in select(self._receiver_1, self._receiver_2):  # (10)!
            if selected_from(selected, self._receiver_1):  # (11)!
                print(f"Received from receiver_1: {selected.value}")
                await self._output.send(selected.value)
                if not selected.value:  # (12)!
                    break
            elif selected_from(selected, self._receiver_2):  # (13)!
                print(f"Received from receiver_2: {selected.value}")
                await self._output.send(selected.value)
                if not selected.value:  # (14)!
                    break
            else:
                assert False, "Unknown selected channel"
        print("EchoActor finished")
    # (15)!


# (3)!
input_channel_1 = Broadcast[bool]("input_channel_1")
input_channel_2 = Broadcast[bool]("input_channel_2")
echo_channel = Broadcast[bool]("echo_channel")

echo_actor = EchoActor(  # (4)!
    input_channel_1.new_receiver(),
    input_channel_2.new_receiver(),
    echo_channel.new_sender(),
    "echo-actor",
)

echo_receiver = echo_channel.new_receiver()  # (5)!

async def main() -> None:  # (6)!
    # (8)!
    await input_channel_1.new_sender().send(True)
    await input_channel_2.new_sender().send(False)

    await run(echo_actor)  # (9)!

    await echo_channel.close()  # (16)!

    async for message in echo_receiver:  # (17)!
        print(f"Received {message=}")


if __name__ == "__main__":  # (7)!
    asyncio.run(main())
```

1. We define an `EchoActor` that receives messages from two channels and sends
    them to another channel.

2. We implement the [`_run()`][_run] method that will receive messages from the two
    channels using [`select()`][frequenz.channels.select] and send them to the
    output channel. The `run()` method will stop if a `False` message is received.

3. We create the channels that will be used with the actor.

4. We create the actor and connect it to the channels by creating new receivers and
    senders from the channels.

5. We create a receiver for the `echo_channel` to eventually receive the messages sent
    by the actor.

6. We define the `main()` function that will run the actor.

7. We start the `main()` function in the async loop using [`asyncio.run()`][asyncio.run].

8. We send a message to each of the input channels. These messages will be queued in
    the channels until they are consumed by the actor.

9. We start the actor and wait for it to finish using the
    [`run()`][frequenz.sdk.actor.run] function.

10. The [`select()`][frequenz.channels.select] function will get the first message
    available from the two channels. The order in which they will be handled is
    unknown, but in this example we assume that the first message will be from
    `input_channel_1` (`True`) and the second from `input_channel_1` (`False`).

11. The [`selected_from()`][frequenz.channels.selected_from] function will return
    `True` for the `input_channel_1` receiver. `selected.value` holds the received
    message, so `"Received from receiver_1: True"` will be printed and `True` will be
    sent to the `output` channel.

12. Since `selected.value` is `True`, the loop will continue, going back to the
    [`select()`][frequenz.channels.select] function.

13. The [`selected_from()`][frequenz.channels.selected_from] function will return
    `False` for the `input_channel_1` receiver and `True` for the `input_channel_2`
    receiver. The message stored in `selected.value` will now be `False`, so
    `"Received from receiver_2: False"` will be printed and `False` will be sent to the
    `output` channel.

14. Since `selected.value` is `False`, the loop will break.

15. The [`_run()`][_run] method will finish normally and the actor will be stopped, so
    the [`run()`][frequenz.sdk.actor.run] function will return.

16. We close the `echo_channel` to make sure the `echo_receiver` will stop receiving
    messages after all the queued messages are consumed (otherwise the step 17 will
    never end!).

17. We receive the messages sent by the actor to the `echo_channel` one by one and print
    them, it should print first `Received message=True` and then `Received
    message=False`.

The expected output is:

```
Received from receiver_1: True
Received from receiver_2: False
Received message=True
Received message=False
```

[async context manager]: https://docs.python.org/3/reference/datamodel.html#async-context-managers
[_run]: #the-_run-method
"""

from ..timeseries._resampling import ResamplerConfig
from ._actor import Actor
from ._background_service import BackgroundService
from ._run_utils import run

__all__ = [
    "Actor",
    "BackgroundService",
    "ResamplerConfig",
    "run",
]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Utilities for checking and clamping bounds and power values to exclusion bounds."""

from frequenz.quantities import Power

from ...timeseries import Bounds


def check_exclusion_bounds_overlap(
    lower_bound: Power,
    upper_bound: Power,
    exclusion_bounds: Bounds[Power] | None,
) -> tuple[bool, bool]:
    """Check if the given bounds overlap with the given exclusion bounds.

    Example:

        ```
                       lower                        upper
                          .----- exclusion zone -----.
        -----|✓✓✓✓✓✓✓✓✓✓✓✓|xxxxxxxxxxxxxxx|----------|----
             `-- usable --'-- exclusion --´
             |                 overlap    |
             |                            |
           lower                        upper
           bound                        bound
                              (inside the exclusion zone)
        ```

        Resulting in `(False, True)` because only the upper bound is inside the
        exclusion zone.

    Args:
        lower_bound: The lower bound to check.
        upper_bound: The upper bound to check.
        exclusion_bounds: The exclusion bounds to check against.

    Returns:
        A tuple containing a boolean indicating if the lower bound is bounded by the
            exclusion bounds, and a boolean indicating if the upper bound is bounded by
            the exclusion bounds.
    """
    if exclusion_bounds is None:
        return False, False

    bounded_lower = False
    bounded_upper = False

    if exclusion_bounds.lower < lower_bound < exclusion_bounds.upper:
        bounded_lower = True
    if exclusion_bounds.lower < upper_bound < exclusion_bounds.upper:
        bounded_upper = True

    return bounded_lower, bounded_upper


def adjust_exclusion_bounds(
    lower_bound: Power,
    upper_bound: Power,
    exclusion_bounds: Bounds[Power] | None,
) -> tuple[Power, Power]:
    """Adjust the given bounds to exclude the given exclusion bounds.

    Args:
        lower_bound: The lower bound to adjust.
        upper_bound: The upper bound to adjust.
        exclusion_bounds: The exclusion bounds to adjust to.

    Returns:
        The adjusted lower and upper bounds.
    """
    if exclusion_bounds is None:
        return lower_bound, upper_bound

    # If the given bounds are within the exclusion bounds, there's no room to adjust,
    # so return zero.
    #
    # And if the given bounds overlap with the exclusion bounds on one side, then clamp
    # the given bounds on that side.
    match check_exclusion_bounds_overlap(lower_bound, upper_bound, exclusion_bounds):
        case (True, True):
            return Power.zero(), Power.zero()
        case (False, True):
            return lower_bound, exclusion_bounds.lower
        case (True, False):
            return exclusion_bounds.upper, upper_bound
    return lower_bound, upper_bound


# Just 20 lines of code in this function, but unfortunately 8 of those are return
# statements, and that's too many for pylint.
def clamp_to_bounds(  # pylint: disable=too-many-return-statements
    value: Power,
    lower_bound: Power,
    upper_bound: Power,
    exclusion_bounds: Bounds[Power] | None,
) -> tuple[Power | None, Power | None]:
    """Clamp the given value to the given bounds.

    When the given value can falls within the exclusion zone, and can be clamped to
    both sides, both options will be returned.

    When the given value falls outside the usable bounds and can be clamped only to
    one side, only that option will be returned.

    Args:
        value: The value to clamp.
        lower_bound: The lower bound to clamp to.
        upper_bound: The upper bound to clamp to.
        exclusion_bounds: The exclusion bounds to clamp outside of.

    Returns:
        The clamped value.
    """
    # If the given bounds are within the exclusion bounds, return zero.
    #
    # And if the given bounds overlap with the exclusion bounds on one side, and the
    # given power is in that overlap region, clamp it to the exclusion bounds on that
    # side.
    if exclusion_bounds is not None:
        match check_exclusion_bounds_overlap(
            lower_bound, upper_bound, exclusion_bounds
        ):
            case (True, True):
                return None, None
            case (True, False):
                if value < exclusion_bounds.upper:
                    return None, exclusion_bounds.upper
            case (False, True):
                if value > exclusion_bounds.lower:
                    return exclusion_bounds.lower, None

    # If the given value is outside the given bounds, clamp it to the closest bound.
    if value < lower_bound:
        return lower_bound, None
    if value > upper_bound:
        return None, upper_bound

    # If the given value is within the exclusion bounds and the exclusion bounds are
    # within the given bounds, clamp the given value to the closest exclusion bound.
    if exclusion_bounds is not None and not value.isclose(Power.zero()):
        if exclusion_bounds.lower < value < exclusion_bounds.upper:
            return exclusion_bounds.lower, exclusion_bounds.upper

    return value, value
